
#ifndef NETWORKLAYER_DTRB_DTRB_H
#define NETWORKLAYER_DTRB_DTRB_H


#include <omnetpp.h>
#include <algorithm>

#include "IPAddress.h"
#include "IP.h"
#include "IInterfaceTable.h"
#include "InterfaceTableAccess.h"
#include "IRoutingTable.h"
#include "RoutingTableAccess.h"
#include "Ieee802Ctrl_m.h"
#include "ICMPMessage.h"




#include "networklayer/dtrb/DTRB_RouteTable.h"
#include "networklayer/dtrb/DTRB_CTRL_m.h"
#include "networklayer/dtrb/DTRB_RouteBlock.h"
#include "networklayer/dtrb/DTRB_Timeout_m.h"
#include "networklayer/dtrb/DTRB_Timer.h"
#include "networklayer/dtrb/DTRB_Packet_m.h"
#include "networklayer/dtrb/AppPkt_m.h"


using namespace DTRB;

namespace DTRB {

//===========================================================================================
// class QueueElement: defines a data type for all queued data messages
//===========================================================================================
class QueueElement : public cPacket
{
	public:
		QueueElement(){ msgId = 0; msgCreationTime = 0; destAddr = 0; srcAddr = 0; obj = NULL; };

		QueueElement(unsigned int msgId, simtime_t msgCreationTime, unsigned int destAddr, unsigned int srcAddr, DTRB_Packet* obj){
			this->msgId = msgId;
			this->msgCreationTime = msgCreationTime;
			this->destAddr = destAddr;
			this->srcAddr = srcAddr;
			this->obj = obj;
		};
		virtual ~QueueElement(){};
		unsigned int getMsgId(){ return msgId; };
		simtime_t getCreationTime(){ return msgCreationTime; };
		unsigned int getDestAddr(){ return destAddr; };
		unsigned int getSrcAddr(){ return srcAddr; };
		DTRB_Packet* getObject(){ return obj; };

	private:
		unsigned int msgId;
		simtime_t msgCreationTime;
		unsigned int destAddr;
		unsigned int srcAddr;
		DTRB_Packet* obj;

};

}


class DTRB_Net : public cSimpleModule
{
  public:
    int numInitStages() const  {return 4;}
    void initialize(int);
    void finish();


    /** @brief Function called whenever a message arrives at the module */
    void handleMessage(cMessage * msg);

    /** @brief change the address we think we are reachable at. Does not(!) make sure we are actually reachable there. The change request will be ignored if this node has already participated in DTRB */
    void setMyAddr(unsigned int myAddr);

    DTRB_RouteTable* getDTRBRouteTable();

    /** @brief guesses which router the given address belongs to, might return 0 */
	cModule* getRouterByAddress(IPAddress address);

  private:
      //friend class DTRB_RoutingTable;

      //void processPacket (const IPDatagram* datagram);
      //===============================================================================
      // OPERATIONS
      //===============================================================================

      /** @brief Handle self messages such as timer... */
      virtual void handleSelfMsg(cMessage*);

      struct compareEntries {
          bool operator ()(DTRB_RouteEntry *lhs, DTRB_RouteEntry *rhs) { return (lhs->maxQvalue()) > (rhs->maxQvalue()); }
        };

      /** @brief Handle messages from upper layer */
      virtual void handleUpperMsg(cPacket*);
      virtual void handleLowerMsg(cPacket*);

      /** @brief Function handles lower messages: sends messages sent to the node to higher layer, forward messages to destination nodes... */
      virtual void handleLowerCTRL(DTRB_CTRL*); // Group Message Algorithm
      virtual void handleLowerPacket(DTRB_Packet*); // DTN Message List
      void updateRouteBlock(const DTRB_RouteBlock& mb, simtime_t now, unsigned int custodianAddress);
      DTRB_Packet* encapsMsg(cPacket*);
      void sendUp(cPacket*);

      /** @brief Function sends messages to lower layer (transport layer) */
      void sendDown(cPacket*, int);

      /** @brief calculates a backoff value */
      simtime_t computeBackoff(simtime_t);

      void insertQueuedMsg(DTRB_Packet *packet);

      DTRB_Packet* getQueuedMsg(unsigned int destAddr);

      void removeQueuedMsg(unsigned int destAddr);

      /** @brief generates and sends down a new GAM msg */
      void sendCTRL();


      bool checkAndSendQueuedPkts(unsigned int destinationAddress, unsigned int custodianAddress);

      //===============================================================================
      // MEMBER VARIABLES
      //===============================================================================
      /** @brief pointer to the routing table */
      DTRB_RouteTable * dtrb_routeTable;

      DTRB_Timer* ctrlTimeout; // when CTRL msg period expired

      /** @brief address of the node */
      unsigned int myAddr;

      int numQueuedMsgs;
      int numRcvdOldMsgs;
      int numDeletedMsgs;

      /** @brief vector contains the queued data packets, waiting for a route */
      cQueue * queuedElements;

      /** @brief defines the routing block size */
      unsigned int rblockSize;

      int mHeaderLength;

      int totalPacketsSent; /**< total number of packets sent down to lower layer */
      int totalPacketsRcvd; /**< total number of packets sent up to app layer */
      int totalPacketsDeleted; /**< total number of packets received but deleted */
      int totalBytesSent; /**< total number of bytes sent down to lower layer */

      int statsCTRLSent; /**< number of generated DTRB CTRL msg */
      int statsCTRLRcvd; /**< number of consumed DTRB CTRL msg */
      int statsTrafficSent; /**< number of injected Application-layer messages */
      int statsTrafficRcvd; /**< number of received Application-layer messages (for this node) */
      cOutVector statsNodeDensity;

      //int statsDTRBRcvd; /**< number of observed DTRB messages */

      simtime_t dataLatency;
      int dataSamples;
      cOutVector dataDelayVec;

      typedef std::vector<unsigned int> MessageVector;
      MessageVector messageVector;



      int RESPONSIBLE_ADDRESSES_PREFIX; /**< NED parameter: netmask of network this DTRB router is responsible for, -1 for self only */
      const char* DTRB_INTERFACES; /**< NED parameter: list of interfaces to run DTRB on, separated by a single space character */
      IPAddress AUTOASSIGN_ADDRESS_BASE; /**< NED parameter: start of address range from which to automatically assign an address to the DTRB_INTERFACES */
      int CTRL_MSG_PERIOD; /**< NED parameter: GAM msg frequency  */
      int BUFFER_SIZE_PACKETS; /**< NED configuration parameter: maximum number of queued packets, -1 for no limit */
      int BUFFER_SIZE_BYTES; /**< NED configuration parameter: maximum total size of queued packets, -1 for no limit */

      /** @brief gate id*/
		//@{
		int mUppergateIn;
		int mUppergateOut;
		int mLowergateIn[2];
		int mLowergateOut[2];
		//@}

  };

  #endif


