#include "DTRB.h"
#include "UDPControlInfo_m.h"
#include "UDPSocket.h"
#include "IPProtocolId_m.h"
#include "UDPPacket.h"
#include "ModuleAccess.h"
#include "IP.h"
#include "IPControlInfo.h"



Define_Module( DTRB_Net );

namespace {
	const double DELTA = 0.85;
	const double OMEGA = 0.95;
	const int DTRB_CTRL_HEADER_LENGTH = 13; /**< length (in bytes) of a DTRB RM header */
	const int DTRB_RBLOCK_LENGTH = 10; /**< length (in bytes) of one DTRB RBlock */
	//const int UDPPort = 9000; /**< UDP Port to listen on (TBD) */
	const double MAXJITTER = 0.001; /**< all messages sent to a lower layer are delayed by 0..MAXJITTER seconds (draft-ietf-manet-jitter-01) */
	const IPAddress LL_MANET_ROUTERS = "224.0.0.90"; /**< Link-local multi-cast address of all MANET routers (TBD) */
	//const int INFINITO = 1000;
}

void DTRB_Net::initialize(int aStage)
{
    cSimpleModule::initialize(aStage);

    if (0 == aStage)
    {

    	mUppergateIn  = findGate("uppergate$i");
		mUppergateOut = findGate("uppergate$o");
		mLowergateIn[0]  = findGate("ifIn", 0);
		mLowergateOut[0] = findGate("ifOut", 0);
		mLowergateIn[1]  = findGate("ifIn", 1);
		mLowergateOut[1] = findGate("ifOut", 1);


    	//NED File Parameters
		RESPONSIBLE_ADDRESSES_PREFIX=par("RESPONSIBLE_ADDRESSES_PREFIX");
		DTRB_INTERFACES=par("DTRB_INTERFACES");
		AUTOASSIGN_ADDRESS_BASE=IPAddress(par("AUTOASSIGN_ADDRESS_BASE").stringValue());
		BUFFER_SIZE_PACKETS=par("BUFFER_SIZE_PACKETS");
		BUFFER_SIZE_BYTES=par("BUFFER_SIZE_BYTES");
		CTRL_MSG_PERIOD=par("CTRL_MSG_PERIOD");


    	ctrlTimeout = new DTRB_Timer(this, "ctrlMsgTimeout", CTRL_MSG_PERIOD);
    	WATCH_PTR(ctrlTimeout);

    	//Initialize DTRB stats
    	totalPacketsRcvd = 0;
    	WATCH(totalPacketsRcvd);
        totalPacketsSent = 0;
        WATCH(totalPacketsSent);
        totalBytesSent = 0;
        WATCH(totalBytesSent);
        totalPacketsDeleted = 0;
        WATCH(totalPacketsDeleted);


        statsCTRLSent = 0; /**< number of generated DTRB CTRL MSG */


        statsCTRLRcvd = 0; /**< number of consumed DTRB CTRL MSG */


        dataLatency = 0;
        dataSamples = 0;

        //rblockSize = 13;

        statsNodeDensity.setName("Node Density");

        dataDelayVec.setName("Data delay");
        //dataLoadVec.setName("Data load");
        //controlLoadVec.setName("Control load");

        // -v- DTRBnet needs AUTOASSIGN_ADDRESS_BASE to be set to 0.0.0.0
		if (AUTOASSIGN_ADDRESS_BASE.getInt() != 0) error("DTRBnet needs AUTOASSIGN_ADDRESS_BASE to be set to 0.0.0.0");
		// -^-

		myAddr = AUTOASSIGN_ADDRESS_BASE.getInt() + uint32(getParentModule()->getId());
		queuedElements = new cQueue();
		numQueuedMsgs = 0;
		numRcvdOldMsgs = 0;
		numDeletedMsgs = 0;

       	}

       	if (2 == aStage) {
       		ctrlTimeout->start(CTRL_MSG_PERIOD);
       		dtrb_routeTable = new DTRB_RouteTable(getParentModule(), IPAddress(myAddr), DTRB_INTERFACES, LL_MANET_ROUTERS);
       		WATCH_PTR(dtrb_routeTable);
       	}
}




void DTRB_Net::finish()
{
	recordScalar("totalPacketsRcvd", totalPacketsRcvd);
    recordScalar("totalPacketsSent", totalPacketsSent);
    recordScalar("totalPacketsDeleted", totalPacketsDeleted);
    recordScalar("totalBytesSent", totalBytesSent);
    recordScalar("NumQueuedMsgs", numQueuedMsgs);

    recordScalar("DTRB_CTRLSent", statsCTRLSent);


    recordScalar("DTRB_CTRLRcvd", statsCTRLRcvd);


    recordScalar("NumRcvdOldMsgsd", numRcvdOldMsgs);

    recordScalar("NumDeletedMsgsFromBuffer", numDeletedMsgs);



    if (dataLatency > 0 && dataSamples > 0)
        recordScalar("data latency", dataLatency/dataSamples);


	while (! queuedElements->isEmpty()) {
		QueueElement *el = static_cast<QueueElement *>(queuedElements->remove(queuedElements->get(0)));
		delete el->getObject();
		delete el;
	}
	delete queuedElements;

    delete dtrb_routeTable;
    dtrb_routeTable = 0;


}

void DTRB_Net::handleMessage(cMessage* apMsg) {
	if (mUppergateIn == apMsg->getArrivalGateId()) {
		handleUpperMsg(check_and_cast<cPacket*>(apMsg));
	}
	else if (apMsg->isSelfMessage()) {
		handleSelfMsg(apMsg);
	}
	else {
		cPacket* apPkt = check_and_cast<cPacket*>(apMsg);
		handleLowerMsg(apPkt);
	}
}



/* ------- Process Incoming UpperDataMsg ------- */

void DTRB_Net::handleUpperMsg(cPacket* apMsg) {
	DTRB_Packet * pkt = encapsMsg(apMsg);

	if(pkt->getDestAddress() == myAddr)
	{
		sendUp(pkt->decapsulate());
		delete pkt;
		// -^-

		return;
	}

	insertQueuedMsg(pkt);
}




DTRB_Packet * DTRB_Net::encapsMsg(cPacket * apMsg) {

	AppPkt * control_info = check_and_cast<AppPkt *>(apMsg->removeControlInfo());
	DTRB_Packet * dtrbPkt = new DTRB_Packet(apMsg->getName());

	cModule* destinationModule = getParentModule()->getParentModule()->getModuleByRelativePath(control_info->getDestName());
	if(!destinationModule) error("destination module not found");
	//int msgId = (int) control_info->par("msgId");

	/** get the IP-Address of the destination node **/
	unsigned int destAddr = destinationModule->getId();

	dtrbPkt->setDestAddress(destAddr);
	dtrbPkt->setSrcAddress(myAddr);
	dtrbPkt->setCreationTime(simTime());
	dtrbPkt->setMsgId(control_info->getMsgId());
	dtrbPkt->encapsulate(apMsg);
	delete control_info;
	totalPacketsSent++;
	return dtrbPkt;
}



DTRB_Packet* DTRB_Net::getQueuedMsg(unsigned int destAddr){
	cQueue::Iterator queueIterator(*queuedElements);
	QueueElement * qElement;

	while(!queueIterator.end()){
		qElement = dynamic_cast<QueueElement*>(queueIterator());
		if ((qElement) && (qElement->getDestAddr() == destAddr)) {
			return dynamic_cast<DTRB_Packet*>(qElement->getObject());
		}
		queueIterator++;
	}
	return NULL;
}

void DTRB_Net::removeQueuedMsg(unsigned int destAddr){
	cQueue::Iterator queueIterator(*queuedElements);
	QueueElement * qElement;

	while((!queueIterator.end()) && (queuedElements->length() > 0)){
		qElement = dynamic_cast<QueueElement*>(queueIterator());
		if ((qElement) && (qElement->getDestAddr() == destAddr)) {
			delete queuedElements->remove(queueIterator());
			return;
		}
		queueIterator++;
	}
}

void DTRB_Net::insertQueuedMsg(DTRB_Packet *packet){


	cQueue::Iterator queueIterator(*queuedElements);
	QueueElement * qElement1;
	simtime_t oldestMsg = simTime();

	while(!queueIterator.end()){
		qElement1 = dynamic_cast<QueueElement*>(queueIterator());
		if ((qElement1) && (qElement1->getMsgId() == packet->getMsgId())) {
			EV << "Duplicated packet: ";
			delete(packet);
			return;
		}
		if (qElement1->getCreationTime() < oldestMsg)
			oldestMsg = qElement1->getCreationTime();
		queueIterator++;
	}

	//if ((queueIterator.lenght() < BUFFER_SIZE_PACKETS) && ((dataQueueByteSize + queueIterator.lenght()) < BUFFER_SIZE_BYTES)){
	if (numQueuedMsgs < BUFFER_SIZE_PACKETS){
		ev << "Queueing data packet " << packet->getMsgId() << endl;
		QueueElement * qElement = new QueueElement(packet->getMsgId(), packet->getCreationTime(), packet->getDestAddress(), myAddr, packet);
		queuedElements->insert(qElement);
		//dataQueueByteSize += data->getByteLength();
		numQueuedMsgs++;
		return;
	}


	if (oldestMsg > packet->getCreationTime())
		{
			delete(packet);
			ev << "Old Message!!" << endl;
			numRcvdOldMsgs++;
			return;
		} else {
			queueIterator.init(*queuedElements, false);
			while(!queueIterator.end()){
				qElement1 = dynamic_cast<QueueElement*>(queueIterator());
				if ((qElement1) && (qElement1->getCreationTime() == oldestMsg)) {
					ev << "Erasing data packet " << qElement1->getObject() << endl;
					//dataQueueByteSize -= qd.datagram->getByteLength();
					queuedElements->remove(qElement1);
					numQueuedMsgs--;
					numDeletedMsgs++;
					break;
				}
				queueIterator++;
			}
		}
	//TODO: If necessary, eliminate more than one message (when the simulation has different size of messages). And: numQueuedMsgs statistics.


	ev << "Queueing data packet " << packet << endl;
	QueueElement * qElement = new QueueElement(packet->getMsgId(), packet->getCreationTime(), packet->getDestAddress(), myAddr, packet);
	queuedElements->insert(qElement);
	//dataQueueByteSize += data->getByteLength();
	numQueuedMsgs++;

}


void DTRB_Net::handleLowerMsg(cPacket* apMsg) {
	// detach and free associated ControlInfo object
	delete apMsg->removeControlInfo();

	/**
	 * check the type of received message
	 1) Routing Message: RREQ or RREP
	 2) Error Message: RERR
	 3) Unsupported Message: UERR
	 4) Data Message
	 **/
	if(dynamic_cast<DTRB_CTRL*>(apMsg)) handleLowerCTRL(dynamic_cast<DTRB_CTRL*>(apMsg));
	else if(dynamic_cast<DTRB_Packet*>(apMsg)) handleLowerPacket(dynamic_cast<DTRB_Packet*>(apMsg));
	else error("cannot cast message to DTRB Packet");
}

void DTRB_Net::handleLowerPacket(DTRB_Packet *datagram) {


	IPAddress srcAddr = datagram->getSrcAddress();
	IPAddress destAddr = datagram->getDestAddress();

	ev << "inbound data packet signalled:" << "Source:" << datagram->getSrcAddress() << "Dest.:" << datagram->getDestAddress() << endl;

	// if this is for us, record statistics
	if (destAddr.getInt() == myAddr) {
		ev << "Message is for me!!" << endl;

		std::vector<unsigned int>::iterator iter;

		for (iter = messageVector.begin(); iter < messageVector.end(); iter++)
		{
			if (*iter == datagram->getMsgId())
			{
				EV << "Duplicated packet: ";
				delete datagram;
				totalPacketsDeleted++;
				return;
			}
		}

		messageVector.push_back(datagram->getMsgId());
	    EV << "Received packet: ";

	    dataDelayVec.record(simTime() - datagram->getCreationTime());
		dataLatency += SIMTIME_DBL(simTime() - datagram->getCreationTime());
		dataSamples++;

		sendUp(datagram->decapsulate());
		delete datagram;
		// -^-

		return;
	} else {
		insertQueuedMsg(datagram);
		return;
	}

	//delete datagram;
	// -^-
}

void DTRB_Net::sendUp(cPacket* apMsg) {
	ev << "sending up " << apMsg << "\n";

	totalPacketsRcvd++;
	send(apMsg, mUppergateOut);
}

/* ------- CTRL Processing ------- */


void DTRB_Net::handleLowerCTRL(DTRB_CTRL *routingMsg)
{
    /** message is a routing message **/
    ev << "received message is a CTRL message" << endl;

    statsCTRLRcvd++;

    /** routing message  preprocessing and updating routes from routing blocks **/

    ev << "starting update routes from routing blocks in the received message" << endl;
	std::vector<DTRB_RouteBlock> additional_nodes = routingMsg->getCtrlMsgTable();
	//std::vector<DTRB_AddressBlock> new_additional_nodes;

	//bool isRREQ = (dynamic_cast<DTRB_RREQ*>(pkt) != 0);
	//uint32_t nextHopAddress = getNextHopAddress(routingMsg);
	//InterfaceEntry* nextHopInterface = getNextHopInterface(routingMsg);

	simtime_t now = simTime();

	DTRB_RouteEntry* entry = dtrb_routeTable->getByNode(IPAddress(routingMsg->getOrigNode()));
	if (entry) {
		entry->msgTime = now;
		entry->distTime = (now - routingMsg->getTimestamp());
	} else if (!entry)
	{
		ev << "adding message entry for " << IPAddress(routingMsg->getOrigNode()) << "now" << now<< endl;
		entry = new DTRB_RouteEntry(this);
		dtrb_routeTable->addRoute(entry);
		entry->msgTime = now;
		entry->distTime = (now - routingMsg->getTimestamp());
		entry->nodeAddress = (IPAddress) routingMsg->getOrigNode();
	}

	//entry->setQvalue((IPAddress)routingMsg->getOrigNode(), 1.0, DELTA);
	//if (checkAndSendQueuedPkts(routingMsg->getOrigNode(), routingMsg->getOrigNode()))
	checkAndSendQueuedPkts(routingMsg->getOrigNode(), routingMsg->getOrigNode());


	/* Node density Calculation:

	int origAddr = routingMsg->getOrigNode().getAddress();
	if (nodeDensity.size() == 0){
		nodeDensity.push_back(origAddr);
		statsNodeDensity++;
	}

	for(int j = 0; j < (int)nodeDensity.size(); j++){
		if(nodeDensity[j] == origAddr)
			break;
		if (j == (int)nodeDensity.size() - 1){
			nodeDensity.push_back(origAddr);
			statsNodeDensity++;
		}

	} */



	for (unsigned int i = 0; i < additional_nodes.size(); i++)
		{
		if (additional_nodes[i].getNode()!=myAddr)
			updateRouteBlock(additional_nodes[i], now, routingMsg->getOrigNode());
		}

	//if (brokenRouteCheck)
		//sendGAM();

	//pkt->setAdditionalNodes(new_additional_nodes);

	delete routingMsg;
	//Cluster algorithm


}

bool DTRB_Net::checkAndSendQueuedPkts(unsigned int destinationAddress, unsigned int custodianAddress)
{
	cQueue::Iterator queueIterator(*queuedElements);
	QueueElement * qElement;
	bool repOK = false;

	while(!queueIterator.end()){
		qElement = dynamic_cast<QueueElement*>(queueIterator());
		if ((qElement) && (qElement->getDestAddr() == destinationAddress)){ //&& (qElement->getLastRepAddr() != custodianAddress)) {
			//qElement->setLastRepAddr(custodianAddress);
			DTRB_Packet *pkt = (qElement->getObject()->dup());
			//DTRB_Packet pkt(*qElement->getObject());
			EV << "Sending packet: ";
			statsTrafficSent++;
			sendDown(pkt, custodianAddress);
			repOK = true;
			//return;
		}
		queueIterator++;
	}
	return repOK;
}



void DTRB_Net::updateRouteBlock(const DTRB_RouteBlock& mb, simtime_t now, unsigned int custodianAddress)
{
    DTRB_RouteEntry* entry = dtrb_routeTable->getByNode(IPAddress(mb.getNode()));

    if (entry){
    	if ((entry->msgTime - entry->distTime) < ((simtime_t)mb.getCtrlMsgTime() - (simtime_t)mb.getDist())){
    		entry->msgTime = now;
    		entry->distTime = (now - (simtime_t)mb.getCtrlMsgTime() + (simtime_t)mb.getDist());
    	}

    } else if (!entry)
	{
		ev << "adding route entry for " << IPAddress(mb.getNode()) << "now" << now<< endl;
		entry = new DTRB_RouteEntry(this);
		dtrb_routeTable->addRoute(entry);
		entry->msgTime = now;
		entry->distTime = (now - (simtime_t)mb.getCtrlMsgTime() + (simtime_t)mb.getDist());
		entry->nodeAddress = (IPAddress) mb.getNode();
	}

    //(checkAndSendQueuedPkts(entry->nodeAddress.getInt(), custodianAddress));
	if (checkAndSendQueuedPkts(entry->nodeAddress.getInt(), custodianAddress))
		entry->setQvalue((IPAddress) custodianAddress, mb.getQvalue(), DELTA);

}

void DTRB_Net::sendCTRL()
{

    //ev << "Preparing GAM broadcast " << destAddr << endl;

    DTRB_CTRL *my_ctrl = new DTRB_CTRL("CTRL");

    my_ctrl->setOrigNode(myAddr);
    my_ctrl->setTimestamp(simTime());

    /*my_gam->setMsgHdrHopLimit(msgHdrHopLimit);
    my_rreq->getTargetNode().setAddress(destAddr);
    if (targetSeqNum != 0) my_rreq->getTargetNode().setSeqNum(targetSeqNum);
    if (targetDist != 0) my_rreq->getTargetNode().setDist(targetDist); */

    //if (RESPONSIBLE_ADDRESSES_PREFIX != -1) my_ctrl->setOrigPrefix(RESPONSIBLE_ADDRESSES_PREFIX);

    std::vector<DTRB_RouteEntry *> RouteVector = dtrb_routeTable->getRouteTable();

    std::sort(RouteVector.begin(), RouteVector.end(), compareEntries());

    for(unsigned int i = 0; i < RouteVector.size(); i++) {
    	DTRB_RouteEntry* entry = RouteVector[i];
    	DTRB_RouteBlock additional_node;
    	additional_node.setCtrlMsgTime(entry->msgTime.dbl());
    	additional_node.setNode(entry->nodeAddress.getInt());
    	additional_node.setDist(SIMTIME_DBL(entry->distTime));
    	additional_node.setQvalue(entry->maxQvalue());
    	//additional_node.setQvalue(entry->distTime);
    	my_ctrl->getCtrlMsgTable().push_back(additional_node);
    }


    sendDown(my_ctrl, LL_MANET_ROUTERS.getInt());
    statsCTRLSent++;
}


/* ------- Data Processing ------- */


void DTRB_Net::handleSelfMsg(cMessage* apMsg)
{
    ev << "handle self message" << endl;

    std::vector<DTRB_RouteEntry *> RouteVector = dtrb_routeTable->getRouteTable();

    int numConnected = 0;

   	for(unsigned int i = 0; i < RouteVector.size(); i++) {
   		DTRB_RouteEntry* entry = RouteVector[i];
   		if ((simTime() - entry->msgTime) < 10) numConnected++;
   	}

   	statsNodeDensity.record(numConnected);

    if (dynamic_cast<DTRB_Timeout*>(apMsg) != NULL)
    {
        // Something timed out. Let's find out what.

        if (ctrlTimeout->isExpired())
        {
        	double mult = pow(OMEGA, 2*CTRL_MSG_PERIOD);
			std::vector<DTRB_RouteEntry *> RouteVector = dtrb_routeTable->getRouteTable();

			for(unsigned int i = 0; i < RouteVector.size(); i++) {
				DTRB_RouteEntry* entry = RouteVector[i];
				entry->ageDeliveryPreds(mult);
			}
        	sendCTRL();
        	ctrlTimeout->start(CTRL_MSG_PERIOD);
        }

    }
    else error("unknown message type");

}

void DTRB_Net::sendDown(cPacket* apMsg, int destAddr) {
	// WARNING: This method uses lots of hacks and/or guesswork to determine output gate and message recipient!
	// More specifically, Hosts are assumed to only have interfaces "wlan", "eth" or both. In the latter case, lowergateOut[1] is expected to be connected to the "eth" interface.

	// all messages sent to a lower layer are delayed by 0..MAXJITTER seconds (draft-ietf-manet-jitter-01)
	simtime_t jitter = dblrand() * MAXJITTER;

	// set byte size of message
	const DTRB_Packet* data = dynamic_cast<const DTRB_Packet*>(apMsg);
	const DTRB_CTRL* ctrl = dynamic_cast<const DTRB_CTRL*>(apMsg);

	if (ctrl) {
		apMsg->setByteLength(DTRB_CTRL_HEADER_LENGTH + (1 + ctrl->getCtrlMsgTable().size() * DTRB_RBLOCK_LENGTH));
	}
	else if (data) {

	}
	else {
		error("tried to send unsupported message type");
	}

	if(destAddr == (int)LL_MANET_ROUTERS.getInt()) {
		MACAddress mac_dest = MACAddress::BROADCAST_ADDRESS;
		apMsg->removeControlInfo();
		Ieee802Ctrl *control_info = new Ieee802Ctrl();
		control_info->setDest(mac_dest);
		apMsg->setControlInfo(control_info);

		if (mLowergateOut[0] != -1) sendDelayed(apMsg, jitter, mLowergateOut[0]);
		if (mLowergateOut[1] != -1) {
			cPacket* apMsg2 = (cPacket*)apMsg->dup();
			apMsg2->removeControlInfo();
			Ieee802Ctrl *control_info2 = new Ieee802Ctrl();
			control_info2->setDest(mac_dest);
			apMsg2->setControlInfo(control_info2);
			sendDelayed(apMsg2, jitter, mLowergateOut[1]);
		}

		return;
	}

	cModule* srcHost = getParentModule();
	if (!srcHost) error("source getModule(i.e. our parent) for new message not found");
	IInterfaceTable* srcIftable = check_and_cast<IInterfaceTable*>(srcHost->getSubmodule("interfaceTable"));
	if (!srcIftable) error("source module for new message has no interface table \"interfaceTable\"");
	cModule* dstHost = simulation.getModule(destAddr);
	if (!dstHost) error("destination module for new message not found");
	IInterfaceTable* dstIftable = check_and_cast<IInterfaceTable*>(dstHost->getSubmodule("interfaceTable"));
	if (!dstIftable) error("destination module for new message has no interface table \"interfaceTable\"");

	InterfaceEntry* srcWlanIf = srcIftable->getInterfaceByName("wlan");
	InterfaceEntry* srcEthIf = srcIftable->getInterfaceByName("eth");
	InterfaceEntry* dstWlanIf = dstIftable->getInterfaceByName("wlan");
	InterfaceEntry* dstEthIf = dstIftable->getInterfaceByName("eth");

	int gateIndex;
	if (srcWlanIf && !srcEthIf) {
		gateIndex = 0;
	}
	else if (!srcWlanIf && srcEthIf) {
		gateIndex = 0;
	}
	else if (srcWlanIf && srcEthIf) {
		if (dstWlanIf && !dstEthIf) {
			gateIndex = 0;
		}
		else if (!dstWlanIf && dstEthIf) {
			gateIndex = 1;
		}
		else if (dstWlanIf && dstEthIf) {
			gateIndex = 1;
		}
		else error("destination host has neither an interface called \"wlan\" nor an interface called \"eth\" registered in its interfaceTable //0");

	}
	else error("source host has neither an interface called \"wlan\" nor an interface called \"eth\" registered in its interfaceTable //1");

	MACAddress mac_dst;
	if (dstWlanIf && !dstEthIf) {
		mac_dst = dstWlanIf->getMacAddress();
	}
	else if (!dstWlanIf && dstEthIf) {
		mac_dst = dstEthIf->getMacAddress();
	}
	else if (dstWlanIf && dstEthIf) {
		if (srcWlanIf && !srcEthIf) {
			mac_dst = dstWlanIf->getMacAddress();
		}
		else if (!srcWlanIf && srcEthIf) {
			mac_dst = dstEthIf->getMacAddress();
		}
		else if (srcWlanIf && srcEthIf) {
			mac_dst = dstEthIf->getMacAddress();
		}
		else error("source host has neither an interface called \"wlan\" nor an interface called \"eth\" registered in its interfaceTable //2");
	}
	else error("destination host has neither an interface called \"wlan\" nor an interface called \"eth\" registered in its interfaceTable //3");

	apMsg->removeControlInfo();
	Ieee802Ctrl* control_info = new Ieee802Ctrl();
	control_info->setDest(mac_dst);
	apMsg->setControlInfo(control_info);

	sendDelayed(apMsg, jitter, mLowergateOut[gateIndex]);
}




simtime_t DTRB_Net::computeBackoff(simtime_t backoff_var)
{
    return backoff_var * 2;
}





void DTRB_Net::setMyAddr(unsigned int myAddr)
{
    /* Check if this node has already participated in DTRB
    if (statsRREQSent || statsRREPSent || statsRERRSent)
    {
        // TODO: Send RERRs, cold-start DTRB, lose sequence number instead?
        ev << "Ignoring IP Address change request. This node has already participated in DTRB." << endl;
        return;
    }*/

    ev << "Now assuming this node is reachable at address " << myAddr << " (was " << this->myAddr << ")" << endl;
    this->myAddr = myAddr;

    // TODO: if IInterfaceTable was autoconfigured, change IP Address there?
}

DTRB_RouteTable* DTRB_Net::getDTRBRouteTable() {
	return dtrb_routeTable;
}

cModule* DTRB_Net::getRouterByAddress(IPAddress address) {
	return dynamic_cast<cModule*>(simulation.getModule(address.getInt() - AUTOASSIGN_ADDRESS_BASE.getInt()));
}

